#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PURPOSE

REFERENCES

REQUIRES

:author: R.K.Garcia <rkgarcia@wisc.edu>
:copyright: 2018 by University of Wisconsin Regents, see AUTHORS for more details
:license: GPLv3, see LICENSE for more details
"""
import os
import sys
import logging
import zlib
import struct
import fcntl
import time
import json
from signal import alarm, signal, SIGALRM, SIGCONT, SIGHUP, SIGUSR1, SIGUSR2, pause
from amqpfind import AmqpExchange
try:
    import cPickle as pkl
except ImportError:
    import pickle as pkl


LOG = logging.getLogger(__name__)

# default location for FIFO to contact our daemon with
FIFO_LOCK_PID_DIR = os.environ.get("AMQPSEND_DIR", os.path.expanduser("~/.amqpsend"))

DEFAULT_FIFO_FMT = "{host:s}-{user:s}-{exchange:s}.fifo"  # used for clients to send serialized data to daemon
DEFAULT_LOCK_FMT = "{host:s}-{user:s}-{exchange:s}.lock"  # locked when daemon is active and available
DEFAULT_PID_FMT = "{host:s}-{user:s}-{exchange:s}.pid"  # locked when client is connected, as a precaution

DEFAULT_OUTGOING_HOST = 'mq1'
DEFAULT_OUTGOING_USER = 'himawari'
DEFAULT_OUTGOING_PASSWORD = 'guest'
DEFAULT_OUTGOING_EXCHANGE = 'satellite'
DEFAULT_OUTGOING_TOPIC = "testing.one.two.three"


class CRCError(ValueError):
    pass


def serialize(obj):
    """prefix a pickle with with CRC32 and length"""
    payload = pkl.dumps(obj, pkl.HIGHEST_PROTOCOL)
    crc = zlib.crc32(payload)
    nbytes = len(payload)
    pid = os.getpid()
    LOG.debug('nbytes={0} pid={1} crc={2}'.format(nbytes, pid, crc))
    prefix = struct.pack('iiq', nbytes, pid, crc)
    return prefix + payload


def deserialize(fob):
    prefix = fob.read(16)
    if len(prefix) == 0:
        raise EOFError("pipe has been closed")
    nbytes, pid, crc = struct.unpack('iiq', prefix)
    payload = fob.read(nbytes)
    if crc != zlib.crc32(payload):
        raise CRCError("CRC32 failed on deserialize")
    return pid, pkl.loads(payload)


class Timeout(Exception):
    pass


class DaemonFailure(Exception):
    """Generated by negotiation failure, or HUP from daemon
    """
    pass


def handle_alarm(*args, **kwargs):
    raise Timeout("timeout alarm fired")


def handle_hup(*args, **kwargs):
    raise DaemonFailure("daemon sent HUP indicating send failure")


USR1 = 0
def handle_usr1(*args, **kwargs):
    global USR1
    USR1 += 1


class AmqpSender(object):
    """Sender Daemon for AMQP using pika

    """
    fifo_path = None
    timeout = None
    host, user, password, exchange = None, None, None, None

    def get(self):
        """Yield a series of (topic:str, content:dict) tuples
        which are deserialized from one or more connection cycles to incoming fifo
        """
        fifo = None
        client_pid = None
        while True:
            if fifo is None:
                LOG.info("connecting to fifo")
                fifo = open(self.fifo_path, 'rb')  # blocks until we have a customer
            try:
                client_pid, obj = deserialize(fifo)  # returns None if nothing more from this customer; ValueError on CRC mismatch
            except CRCError:
                LOG.error("deserialization CRC failure, will close and reconnect")
                obj = None
            except EOFError:  # time to connect to new client
                fifo.close()
                fifo = None
                LOG.debug("closing incoming fifo")
                continue
            for tup in obj:  # expect list of (topic, content) tuples
                yield tup
            yield client_pid

    def __call__(self, *messages):
        """
        """
        amqp = None

        if self.timeout:
            signal(SIGALRM, handle_alarm)
            alarm(self.timeout)

        failures_since_last_pid = 0

        if not messages:  # then we're talking to a daemon
            messages = self.get()
        try:
            for tosend in messages:  # expect series of (topic,content) tuples

                if isinstance(tosend, int):  # then it's a client pid to send ack/nak
                    if failures_since_last_pid > 0:
                        LOG.warning("sending nak (SIGHUP) to client pid {0} due to {1} failures since last connection".format(tosend, failures_since_last_pid))
                        try:
                            os.kill(tosend, SIGHUP)
                        except:
                            pass
                    else:
                        LOG.debug("sending ack (SIGCONT) to client pid {0}".format(tosend))
                        try:
                            os.kill(tosend, SIGCONT)
                        except:
                            pass
                    failures_since_last_pid = 0
                    continue

                if amqp is None:
                    # fetch the commandline option we added in add_options
                    LOG.info('connecting to outgoing server %s' % self.host)
                    try:
                        amqp = AmqpExchange(host=self.host,
                                            user=self.user,
                                            password=self.password,
                                            exchange=self.exchange)
                    except:
                        LOG.critical("unable to connect to %s for outgoing messages" % self.host)
                        amqp = None

                topic, content = tosend
                try:
                    LOG.debug("sending AMQP message %s:%s" % (topic, repr(content)))
                    amqp.produce(content, key=topic)
                    if self.timeout:
                        alarm(self.timeout)
                except:
                    LOG.error("could not send message with topic %s" % topic)
                    amqp = None
                    failures_since_last_pid += 1
        except KeyboardInterrupt:
            LOG.info("keyboard interrupt: that's all, folks!")
        except Timeout:
            LOG.info("timeout - shutting down!")
        if failures_since_last_pid > 0:  # typically only happens in immediate mode, or on timeout
            raise AssertionError("not all messages were successfully sent")

    def __init__(self, host, user, password, exchange, fifo_path=None, timeout=None):
        # start sender thread
        self.fifo_path = fifo_path
        self.timeout = timeout
        self.host, self.user, self.password, self.exchange = host, user, password, exchange


def main_daemon(lock_path, fifo_path, pid_path, gpid, host, user, password, exchange, timeout=None):
    daemon = AmqpSender(host, user, password, exchange, fifo_path, timeout=timeout)
    daemon()
    sys.exit(0)


def pid_is_ours_and_alive(pid):
    from os import kill
    # from os import ProcessLookupError
    try:
        kill(pid, SIGCONT)
    except:
        return False
    return True


def await_exists(path, attempts=5, delay=1.0):
    for i in range(attempts):
        if os.path.exists(path):
            return True
        time.sleep(delay)
    raise EnvironmentError("{0} was expected but never showed up".format(path))


def connect_daemon(host, user, password, exchange, lock_dir=FIFO_LOCK_PID_DIR, timeout=None):
    """Check whether daemon is alive and ready to accept outgoing messages
    If so, return a file object representing the connection to the daemon
    Else create such a daemon and return connection
    """
    lock_filename = DEFAULT_LOCK_FMT.format(**locals())
    fifo_filename = DEFAULT_FIFO_FMT.format(**locals())
    pid_filename = DEFAULT_PID_FMT.format(**locals())

    lock_path, fifo_path, pid_path = (os.path.join(lock_dir, lock_filename),
                                      os.path.join(lock_dir, fifo_filename),
                                      os.path.join(lock_dir, pid_filename))

    if not os.path.isdir(lock_dir):
        LOG.info('creating lock+fifo directory {0}'.format(lock_dir))
        os.makedirs(lock_dir)

    # check lock file; if it's locked we have a pre-existing daemon, else we become the daemon
    lock_fob = open(lock_path, 'wb+')
    try:
        fcntl.flock(lock_fob, fcntl.LOCK_EX | fcntl.LOCK_NB)
        LOG.info("successfully acquired lock; no daemon yet available")
    except IOError:
        # someone already owns the lock, so we can connect
        LOG.debug("lock is owned, will connect to existing or upcoming daemon")
        await_exists(fifo_path)
        await_exists(pid_path)
        pid_fob = open(pid_path, 'r')
        try:
            daemon_pid = int(pid_fob.read().strip())
        except IOError:
            LOG.error('daemon pidfile not readable as {0}'.format(pid_path))
            daemon_pid = None
        except ValueError:
            LOG.warning("could not read int from pidfile")
            daemon_pid = None
        if (daemon_pid is not None) and pid_is_ours_and_alive(daemon_pid):
            fifo_fob = open(fifo_path, 'wb', 0)
            LOG.info("fifo to daemon on pid {0} connected".format(daemon_pid))
            lock_fob.close()
            return fifo_fob, pid_fob
        raise DaemonFailure("daemon was expected but wouldn't negotiate")

    # otherwise we now own the lock, time to daemonize

    if not os.path.exists(fifo_path):
        os.mkfifo(fifo_path)
    gpid = os.getpid()  # note who to signal ready (SIGUSR1)
    signal(SIGUSR1, handle_usr1)
    usr1_base = USR1

    pid = os.fork()
    if pid == 0:  # then we are the child
        pid = os.fork()
        if pid == 0:  # then we are the grandchild
            # signal grandparent that we're ready
            with open(pid_path + '-', 'wb') as pidfob:
                pidfob.write("{0}\n".format(os.getpid()).encode('ASCII'))
            os.rename(pid_path + '-', pid_path)
            LOG.info("pid is {0}, noted in {1}".format(os.getpid(), pid_path))
            fcntl.flock(lock_fob, fcntl.LOCK_EX)  # this is technically redundant since advisory lock is inherited
            LOG.info("daemon has confirmed lock")
            os.kill(gpid, SIGUSR1)
            main_daemon(lock_path, fifo_path, pid_path, gpid, host, user, password, exchange, timeout=timeout)
        sys.exit(0)
    else:  # we are the parent, and we have data to send
        # wait for child to acknowledge ready
        signal(SIGALRM, handle_alarm)
        alarm(5)
        while USR1 != (usr1_base + 1):
            pause()  # await USR1
        LOG.info("handing lock fully over to daemon")
        lock_fob.close()
        LOG.info("daemon is ready")
        try:
            lock_fob = open(lock_path, 'wb+')
            fcntl.flock(lock_fob, fcntl.LOCK_EX | fcntl.LOCK_NB)
            LOG.error('daemon should have taken the lock by this point')
            raise DaemonFailure("daemon never fully came up")
        except IOError:  # not locked
            pass
        lock_fob.close()
        alarm(10)
        fifo_fob = open(fifo_path, 'wb', 0)
        pid_fob = open(pid_path, 'r')
        return fifo_fob, pid_fob


def messages_from_input(stdin=None, topic=None, *json_dicts):
    if json_dicts and topic:
        for json_dict in json_dicts:
            content = json.loads(json_dict)
            yield topic, content
    if stdin is not None:
        messages = json.load(stdin)
        for topic, content in messages:
            yield topic, content


def amqpsend_to_daemon(fifo_fob, pid_fob, messages):
    LOG.info("forwarding {0} messages to daemon".format(len(messages)))
    data = serialize(messages)
    # as an extra precaution, lock the pidfile as a client lock when writing to it to avoid block interleaving
    fcntl.flock(pid_fob, fcntl.LOCK_EX)
    fifo_fob.write(data)
    fifo_fob.close()
    fcntl.flock(pid_fob, fcntl.LOCK_UN)


def amqpsend_immediate(host, user, password, exchange, messages):
    """Send immediately without daemon
    Create an immediate sender,
    read JSON list of (topic:str, content:dict) tuples from stdin,
    forward messages to server before shutting it all down
    """
    LOG.info("sending {0} messages".format(len(messages)))
    sender = AmqpSender(host, user, password, exchange)
    sender(*messages)


def _debug(type_, value, tb):
    """enable with sys.excepthook = debug"""
    if not sys.stdin.isatty():
        sys.__excepthook__(type_, value, tb)
    else:
        import traceback, pdb
        traceback.print_exception(type_, value, tb)
        # …then start the debugger in post-mortem mode.
        pdb.post_mortem(tb)  # more “modern”

DESCRIPTION = r"""Read JSON list of lists [ [topic:str, content:dict], ... ] from stdin and send AMQP message group

Optionally, a single topic and a series of JSON dictionaries can be provided as parameters on command line.
These are sent prior to any messages coming in on stdin. 
Stdin will be skipped if it is a tty (i.e. not a pipe or file redirection).

Example sending one topic, multiple payloads from command line:

./amqpsend.py -vvv -H mq1 -X satellite -u guest -p guest testing.one.two.three '{"foo": "bar"}' '{"baz": "quux"}'

Example sending multiple topics, multiple payloads from stdin
echo '[["testing.one.two.three", {"foo": "bar"}], ["testing.one.two.three", {"baz": "quux"}]]' \
  | ./amqpsend.py -vvv -H mq1 -X satellite -u guest -p guest 

Monitoring the test in another window:
amqpfind -H mq1 -u guest -p guest -X satellite -C 'testing.one.two.three' -j '?' -v

Note: When running in daemon mode, lock, fifo and pid files are in ::FLPD::/ 
This directory is settable as environment AMQPSEND_DIR.

A python-callable client is available as amqpsend.check_call_amqpsend() - see module for documentation.

""".replace('::FLPD::', FIFO_LOCK_PID_DIR)


def main():
    import argparse
    parser = argparse.ArgumentParser(
        description="Send one or more JSON-dictionary AMQP messages in a batch",
        epilog="",
        fromfile_prefix_chars='@')
    # server connection options
    parser.add_argument('-H', '--host', dest='host',
                        help='name of AMQP server to connect to')
    parser.add_argument('-X', '--exchange', dest='exchange',
                        help='name of exchange to connect to')
    parser.add_argument('-u', '--user', dest='user',
                        help='user id to talk to AMQP exchange as')
    parser.add_argument('-p', '--passwd', dest='password',
                        help='password for user')
    parser.add_argument('-D', '--daemon', dest='daemon', type=int, default=None,
                        help='use a background daemon with N seconds inactivity timeout, to consolidate connection; 0 seconds for no timeout')
    # parser.add_argument('-T', '--timeout', dest='timeout', type=int,
    #                     help='seconds to keep daemon connection alive')
    parser.add_argument('-v', '--verbose', dest='verbosity', action="count", default=0,
                        help='each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG')
    parser.add_argument('-d', '--debug', dest='debug', action='store_true',
                        help="enable interactive PDB debugger on exception")
    parser.add_argument('inputs', nargs='*',
                        help="optional input topic followed by json dictionary strings")
    args = parser.parse_args()

    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level=levels[min(3, args.verbosity)],
                        datefmt='%Y-%m-%dT%H:%M:%S',
                        format='%(asctime)s %(levelname)s pid%(process)d:%(module)s:%(funcName)s:L%(lineno)d %(message)s')

    if args.debug:
        sys.excepthook = _debug

    stdin = sys.stdin if not os.isatty(sys.stdin.fileno()) else None
    messages = list(messages_from_input(stdin, *args.inputs))
    if messages:
        LOG.debug('messages to send: {0}'.format(repr(messages)))
    else:
        print(DESCRIPTION)
        return 1
    if args.daemon is not None:
        # obtain a daemon connection
        try:
            fifo, pid = connect_daemon(args.host, args.user, args.password, args.exchange, timeout=args.daemon)

            ackmut = set()
            def ack(*args, **kwargs):
                ackmut.add('done')

            signal(SIGCONT, ack)  # daemon sends this as ACK
            signal(SIGHUP, handle_hup)  # daemon sends this as NAK
            # signal(SIGALRM, handle_alarm)  # already done in connect_daemon
            alarm(16)  # precaution in case of major lock contention or lack of response

            amqpsend_to_daemon(fifo, pid, messages)
            while 'done' not in ackmut:
                pause()
            if 'done' in ackmut:
                LOG.info("received acknowledgement of send from daemon")
                return 0
            # at this point we should have either had ack, timeout, or failure, so this is an error and not a warning
            LOG.error("no timely acknowledgment received from daemon, this should be unreachable state; dropping to immediate send")
        except (Timeout, DaemonFailure):
            LOG.warning("unable to send via daemon, dropping to immediate send")
    amqpsend_immediate(args.host, args.user, args.password, args.exchange, messages)
    return 0


def check_call_amqpsend(cmdline_args, messages):
    """Call amqpsend using subprocess - for use with pythonic clients
    from amqpsend import check_call_amqpsend as amqpsend
    messages = [("testing.one.two.three", {"foo": "bar"}), ("testing.one.two.three", {"baz": "quux"})]
    amqpsend(['-H', 'localhost', '-X', 'satellite', '-u', 'guest', '-p', 'guest', '-vvv', '-D10'], messages)
    """
    import subprocess as sp
    LOG.debug("starting {0}".format(__file__))
    pid = sp.Popen([__file__] + list(cmdline_args), stdin=sp.PIPE)
    LOG.debug("queuing {0} messages".format(len(messages)))
    payload = json.dumps(messages)
    pid.stdin.write(payload.encode("UTF-8"))
    pid.stdin.close()
    LOG.debug("waiting for amqpsend process to complete")
    pid.wait()
    rc = pid.returncode
    if 0 != rc:
        raise sp.CalledProcessError("amqpsend subprocess returncode={0}".format(rc))
    LOG.debug("sent {0} messages".format(len(messages)))


if __name__ == '__main__':
    sys.exit(main())
