import logging
from subprocess import check_call, CalledProcessError
from threading import Thread, active_count, enumerate as active_threads
from amqpfind import main, main_options, AmqpExchange, Queue
import multiprocessing as mp

LOG = logging.getLogger(__name__)

OUTGOING_HOST = 'mq1'
OUTGOING_USER = 'himawari'
OUTGOING_PASSWORD = 'guest'
OUTGOING_EXCHANGE = 'satellite'
OUTGOING_TOPIC = "testing.one.two.three"
PROCESSING_SCRIPT = '/bin/echo'

class ReceiveProcessSend(object):
    """example scaffold for threaded processing of data indicated by AMQP messages, with generation of outgoing AMQP messages on completion
    Receive AMQP messages indicated by command line operations
    Create a new thread, warning if too many threads are active
        Execute external processing command
        Queue a message back to the exchange with arbitrary topic and content
    """
    _amqp_thread = None  # Thread object that's sending AMQP messages which it receives via queue:
    _outgoing_queue = None  # threads, once completed, send tuples of (topic: str, content: dict) to this queue using .send()
    _tasks = None  # list of launched Thread objects that we periodically check on

    def outgoing_sender_main(self):
        """background thread which owns the outgoing AmqpExchange
        This is necessary because pika requires it
        http://pika.readthedocs.io/en/0.10.0/faq.html
        read topic,content tuples from outgoing_queue
        send them to exchange
        exit when outgoing_queue returns a None
        """
        amqp = None
        try:
            while True:
                tosend = self._outgoing_queue.get()
                if tosend is None:  # signal to exit
                    break

                if amqp is None:
                    # fetch the commandline option we added in add_options
                    cmdline_opts = main_options()
                    host = cmdline_opts.outgoing or OUTGOING_HOST
                    LOG.info('connecting to outgoing server %s' % host)
                    try:
                        amqp = AmqpExchange(host=host,
                                            user=OUTGOING_USER,
                                            password=OUTGOING_PASSWORD,
                                            exchange=OUTGOING_EXCHANGE)
                    except:
                        LOG.critical("unable to connect to %s for outgoing messages" % OUTGOING_HOST)
                        amqp = None

                topic, content = tosend
                try:
                    LOG.info("sending AMQP message %s:%s" % (topic, repr(content)))
                    amqp.produce(content, key=topic)
                except:
                    LOG.error("could not send message with topic %s" % topic)
                    amqp = None
        except KeyboardInterrupt as thatsallfolks:
            LOG.info("keyboard interrupt: that's all, folks!")

    def __init__(self, max_tasks=None):
        # start sender thread
        self._tasks = []
        self._max_tasks = max_tasks
        self._outgoing_queue = Queue.Queue()  # inter-thread queue of (topic, content) tuples
        self._amqp_thread = Thread(target=self.outgoing_sender_main, args=())
        self._amqp_thread.daemon = True
        self._amqp_thread.start()

    def send(self, topic, content):
        self._outgoing_queue.put((topic, content))


    def process_main(self, topic, content):
        """thread function to process data and queue messages for outgoing
        IMPLEMENT YOUR OWN FUNCTION HERE, preferably by creating a subclass of ReceiveProcessSend or direct substitution
        """
        # synchronously run the processing script, bombing out thread if processing fails
        cmd = [PROCESSING_SCRIPT, content['path']]
        try:
            check_call(cmd)
        except CalledProcessError as failed:
            LOG.error("processing script %s failed, will not send downstream message" % repr(cmd))
            return

        # queue an outgoing message
        self.send(OUTGOING_TOPIC, {'done': content['path']})

        # perform any cleanup

    def _cleanup(self):
        new_active = []
        for task in self._tasks:
            if task.is_alive():
                new_active.append(task)
            else:
                LOG.debug('cleaning up finished task')
                task.join(0)
        LOG.debug("%d active tasks remain" % len(new_active))
        self._tasks = new_active

    def decline_task(self, topic, content):
        """Handle any logging or other recording of content we're not allowed to task due to resources
        """
        LOG.error("too many tasks! declining %s" % (repr(content)))

    def accept_task(self, topic, content):
        thread = Thread(target=self.process_main, args=(topic, content))
        thread.daemon = True  # allow program to exit without joining this thread
        self._tasks.append(thread)
        thread.start()

    def __call__(self, topic, content):
        """Callback used by amqpfind itself to indicate action that needs to happen
        :param topic:
        :param content:
        :return:
        """
        if self._max_tasks is None:
            self._max_tasks = main_options().tasks
        # if we're out of tasks, see if any cleanup can be done first
        if len(self._tasks) >= self._max_tasks:
            LOG.warning("task load is full, performaing cleanup pass before attempting new task")
            self._cleanup()
        if len(self._tasks) >= self._max_tasks:
            self.decline_task(topic, content)
        else:
            self.accept_task(topic, content)
        self._cleanup()

def my_options(parser):
    """Add any options we need to the command line parser
    """
    parser.add_option('-O', '--outgoing-server', dest='outgoing', default=None,
                        help='outgoing server to queue messages to')
    parser.add_option('-m', '--max-tasks', dest='tasks', type="int", default=mp.cpu_count(),
                        help='maximum number of simultaneous tasks to allow')


# example of subclassing! adopt all functionality of ReceiveProcessSend but provide an alternate process_main

class MyReceiveProcessSend(ReceiveProcessSend):

    def process_main(self, topic, content):
        # synchronously run the processing script, bombing out thread if processing fails
        cmd = [PROCESSING_SCRIPT, "test", content['path']]
        try:
            check_call(cmd)
        except CalledProcessError as failed:
            LOG.error("processing script %s failed, will not send downstream message" % repr(cmd))
            return

        # queue an outgoing message
        self.send(OUTGOING_TOPIC, {'done': content['path']})
        self.send(OUTGOING_TOPIC, {'toast': 'up'})
        LOG.info('queued two messages')


# run amqpfind with new function "process_and_send" available
# python receive_process_send.py -H localhost -X satellite -u himawari -p guest -C 'geo.*.*.*.*.*.*.*.*' --callback process_and_send -vv -O localhost
# monitor outgoing messages with
# python amqpfind.py -H localhost -X satellite -u himawari -p guest -C 'testing.one.two.three' -j '?' -vv
if __name__=='__main__':
    my_processor = MyReceiveProcessSend()  # could also use ReceiveProcessSend base class
    # example of adding options
    main(add_options=my_options, process_and_send=my_processor)
